# pylint: disable=broad-except

# module-level docstring
__doc__='''
Shell command execution
=======================

Sometimes we just want to execute a shell command and possibly
retrieve stdout/stderr, without hassle.
'''

from typing import Iterable, Dict, Union
from subprocess import Popen, PIPE

def execute( command: Union[str,Iterable[str]], shell: bool = True ) -> Dict[str,str]:
    ''' Passes command to subprocess.Popen, retrieves stdout/stderr and performs
    error management.
    Returns a dictionnary containing stdX.
    Upon command failure, prints exception and returns empty dict. '''

    try:
        with Popen( command, stdout=PIPE, stderr=PIPE, shell=shell ) as process:
            # wait and retrieve stdout/err
            _stdout, _stderr = process.communicate()
            # handle text encoding issues and return stdX
            return {
                'stdout': _stdout.decode('utf8', errors='backslashreplace'),
                'stderr': _stderr.decode('utf8', errors='backslashreplace')
            }
    except Exception as e:
        print(f"execute: Error while executing command '{command}' : {e}")
        raise



# module-level docstring
__doc__='''
Debugging utils
===============

Deubgging is made easier with these convenience functions.

API:
    - func_attribute_printout
    - debug_var
    - call_progress

See elements' docstrings for further explanations.
'''

from typing import Callable, Any, Union
import functools
import re
import shutil
import traceback

FUNC_RELEVANT_ATTR={
    '__annotations__',
    '__class__',
    '__defaults__',
    '__dict__',
    '__doc__',
    '__kwdefaults__',
    '__module__',
    '__name__',
    '__qualname__'
}

def func_attribute_printout( user_funtion: Callable ) -> None:
    ''' Prints relevant attributes of `user_function`, with attribute name, value and type.
    Relevant attributes are in `FUNC_RELEVANT_ATTR`.

    Usage example: reviewing an unfamiliar imported function::
    
        from some.module import foo
        func_attribute_printout( foo )
    
    Output::

        foo.__annotations__ = ...
        [...]
        foo.__qualname__ = foo, type=str
        
    '''
    f_name = user_funtion.__name__
    for attr in dir(user_funtion):
        if attr in FUNC_RELEVANT_ATTR:
            attr_val = user_funtion.__getattribute__(attr)
            print(f"{f_name}.{attr} = {attr_val}, type={type(attr_val)}")


regex_var_name_from_call = re.compile(r'\(\s*(?:var\s*=\s*)?([^,\ ]+).*\).*$')

def debug_var( var: Any, var_name: str = None ) -> None:
    ''' Prints the name, value and type of a variable, typically used during development
    for quick and easy type/value sanity checks.

    Warning: if debug_var is decorated, automatic variable name retrieval doesn't work !
             You will need to specify var_name.

    Automatic variable name retrieval : in most situations `debug_var` should be able to 
    retrieve the name of the variable using the `traceback` module. If it fails, simply
    add it manually. This feature was inspired by code snippets from 
    https://stackoverflow.com/questions/2749796/how-to-get-the-original-variable-name-of-variable-passed-to-a-function

    Usage example : you want to add a sanity check to a received value::
    
        ...
        res = do_something()
        debug_var( res )
        ...

    Output::
    
        >>> DEBUG VAR: res=[1999, 2011] (list)

    Note : during development a different method for retrieving var_name automatically
    was found. It used the stack module. Unfortunately it was found to be
    significantly slower and to not have any significant advantage so it was removed.
    You can find the implementation here below::
    
        # inspect method : similar but extract locals from frame(s) and try to find
        stack = inspect.stack()[2:]
        for frameinfos in stack:
            lcls = frameinfos.frame.f_locals
            for name in lcls:
                if id(var) == id(lcls[name]):
                    var_name = name
                    break
            if var_name:
                break
    

    Performance : tested on a AMD 1700 using the standard Python 3.9 interpreter,
    which used only 1 execution thread. Executed 10 iteration in ~2ms and 100'000
    iterations in ~11.8s (~9 iter/ms). Figures for indicative purposes only.
    '''

    # If `var_name` isn't specified, we try to retrieve this information
    if var_name is None:
        # traceback method : extract stack frame to retrieve call code ..
        raw_call_traceback = traceback.extract_stack( limit=2 )[0]
        call_code_line = raw_call_traceback.line
        # .. then extract variable name from call to `debug_var` using regular expression (handles both args and kwargs)
        var_name = regex_var_name_from_call.search(call_code_line).groups()[0]

    if not var_name:
        var_name = "error:Couldn't determine variable name automatically. Pleasy try to pass variable name as argument."
    print( f"DEBUG VAR: {var_name}={var} ({type(var)})")


def call_progress( expected_argument: Union[Callable,str] ) -> Callable:
    ''' Decorator. Typically used to make the execution status of 
    a function verbose when developping.
    The decorated function's execution is enclosed in a text-based
    box, with pre/post-execution message. 

    `expected_argument`: Optional string argument, replaces the default
    message (user_function.__name__).

    Usage example: decorating debug_var (see warning in debug_var's docstring)::
    
        decorated_debug_var = call_progress("debug_var debugs a variable")(debug_var)
        hello='world'
        decorated_debug_var( var=hello, var_name='hello' )

    Output::

        ----------------------------------------
        debug_var debugs a variable ..
        DEBUG VAR: hello=world (<class 'str'>)
        debug_var debugs a variable done
        ----------------------------------------
    
    
    Note: this example is not great: decorating manually (not with @call_progress syntax)
    and decorating a function that specifically advises against decoration. But this showcases
    an actual use case that works.
    '''

    def actual_decorator( user_function: Callable ) -> Callable:

        @functools.wraps( user_function )
        def wrapper( *args, **kwargs ) -> Any:
            nonlocal user_function

            box_W = min( shutil.get_terminal_size().columns, 40 )
            
            print( '-' * box_W )
            print( printout_text + ' ..' )
            res = user_function( *args, **kwargs )
            print( printout_text + ' done' )
            print( '-' * box_W )

            return res

        return wrapper

    if callable( expected_argument ):
        # decorator_with_arguments was run without argument => use 
        # default values for expected arguments or raise error
        user_function = expected_argument
        printout_text = user_function.__name__
        return actual_decorator(user_function)

    if isinstance(expected_argument, str):
        printout_text = expected_argument
        return actual_decorator

    raise ValueError(f"call_progress: `expected_argument` of type {type(expected_argument)} is not in Union[Callable,str] !")



# module-level docstring
__doc__='''
General-purpose decorators
==========================

This file contains useful (function) decorators !

Also, you can find :
 - a structure convention for all function decorators
 - tests for each decorator (see at the bottom of the file; run the file to execute them)

'''

from typing import Any, Callable
import time
import datetime
import logging
from pathlib import Path
import pickle

# functool.wraps is a decorator for wrappers that copies informations from 
# the original function to the wrapper, allowing for transparently chaining decorators
import functools
# for profiling
import cProfile
import pstats 


########### CONVENTION ###########
# Any decorator in this file must 
# follow this format; please use it
# as a template :
########### CONVENTION ###########

def decorator_without_argument( user_function: Callable ) -> Callable:
    '''
    DOCSTRING
    '''

    # maybe do something here

    @functools.wraps( user_function )
    def wrapper( *args, **kwargs ) -> Any:
        nonlocal user_function

        # maybe do something before calling f
        res = user_function( *args, **kwargs )
        # maybe do something after calling f
        return res

    return wrapper
    

def decorator_with_arguments( expected_argument: Any ) -> Callable:
    '''
    DOCSTRING
    '''

    def actual_decorator( user_function: Callable ) -> Callable:

        # maybe do something here

        @functools.wraps( user_function )
        def wrapper( *args, **kwargs ) -> Any:
            nonlocal user_function
            
            # maybe do something before calling f
            res = user_function( *args, **kwargs )
            # maybe do something after calling f

            return res

        return wrapper

    if callable( expected_argument ):
        # decorator_with_arguments was run without argument => use 
        # default values for expected arguments or raise error
        user_function = expected_argument
        return actual_decorator(user_function)

    # decorator_with_arguments was run with argument (maybe do something here)

    return actual_decorator


########### timing ###########

def timer( user_function: Callable ) -> Callable:
    '''
    Prints how much time user function took to run
    '''

    @functools.wraps( user_function )
    def wrapper( *args, **kwargs ) -> Any:
        nonlocal user_function

        t_start = time.time()
        res = user_function( *args, **kwargs )
        elapsed_ms = int( 1000 * (time.time() - t_start) )
        print(f"Function {user_function.__name__} ran for {elapsed_ms}ms.")
        return res

    return wrapper


def profile( user_function: Callable ) -> Callable:
    ''' Profiling is an advanced technique to measure
    code execution time. 

    This decorator was heavily inspired from mCoding's video:
    https://www.youtube.com/watch?v=m_a0fN48Alw

    Using PROF output file : you need the `snakeviz` package
    installed (`pip install snakeviz`). Simply type
    `snakeviz <profile file>`
    in a terminal and you should see a web browser window with 
    visualizations of the profile.
    '''

    @functools.wraps( user_function )
    def wrapper( *args, **kwargs ) -> Any:
        nonlocal user_function

        # Create profiler and run user function
        with cProfile.Profile() as _profile:
            res = user_function( *args, **kwargs )
        
            # Store results
            stats = pstats.Stats( _profile )
            stats.sort_stats( pstats.SortKey.TIME )
            stats.dump_stats(filename=f'profile_{user_function.__name__}.prof')

        return res

    return wrapper


########### logging ###########


def log_to_file( user_function: Callable ) -> Callable:
    '''
    Logs function call to file
    '''
    
    logging.basicConfig( 
        filename=f"{user_function.__name__}.log",
        level=logging.INFO
    )
    _log = logging.getLogger( user_function.__name__ )

    @functools.wraps( user_function )
    def wrapper( *args, **kwargs ) -> Any:
        nonlocal user_function

        str_args = ', '.join([
            str(arg) 
            for arg in args
        ])
        _now = datetime.datetime.now()
        res = user_function( *args, **kwargs )

        _log.info(
            "Called function %s with args=[%s] and kwargs=%s at %s, with return value %s.", 
            user_function.__name__,
            str_args,
            kwargs,
            _now,
            res
        )
        
        return res

    return wrapper


########### caching ###########


def cacheFS( cache_file: Path ) -> Callable:
    '''
    Returns a decorator that caches the returned value of user function using a cache file.
    '''

    def actual_decorator( user_function: Callable ) -> Callable:
        ''' Caches the returned value of user function using a cache file. '''

        @functools.wraps( user_function )
        def wrapper( *args, **kwargs ) -> Any:
            nonlocal user_function, cache_file, cached_data
            
            k = (args, frozenset(kwargs.items()))
            if k in cached_data:
                # cache hit
                print(f"Cache hit for {user_function.__name__} with args={args} and kwargs={kwargs}.")
                return cached_data[k]
            
            # else: cache miss
            res = user_function( *args, **kwargs )
            cached_data[k] = res
            with cache_file.open( mode='wb' ) as f:
                pickle.dump( cached_data, f ) # update cache
            
            print(f"Written cache to file '{cache_file}'.")

            return res

        return wrapper


    def load_cached_data( cache_file: Path ) -> dict:
        if cache_file.is_file():
            try:
                with cache_file.open( mode='rb' ) as f:
                    cached_data = pickle.load( f )
            except EOFError:
                pass # cache file exists but doesn't contain anything/valid data
            else:
                return cached_data
        return dict()


    if callable( cache_file ):
        # cacheFS was run without argument => default filename
        user_function = cache_file
        cache_file = Path( f"{user_function.__name__}.cacheFS" )
        cached_data = load_cached_data( cache_file )
        return actual_decorator(user_function)

    # else: cacheFS was run with argument
    # File Extension enforcement
    if cache_file.suffix != '.cacheFS':
        cache_file = Path( f'{cache_file}.cacheFS' )

    cached_data = load_cached_data( cache_file )

    return actual_decorator


if __name__=="__main__":

    def vspace():
        ''' just prints newlines '''
        print( '\n' * 2 )

    # Tests

    # Test 01
    vspace()
    print("Test 01 : decorator 'timer'")

    @timer
    def wait2s() -> None:
        ''' Just wait 2s '''
        time.sleep(2)

    print(wait2s)
    print(wait2s.__doc__)
    wait2s()

    # Test 02
    vspace()
    print("Test 02 : decorator 'log_to_file'")

    @log_to_file
    def add( a: int, b: int ) -> int:
        ''' Add a and b '''
        return a + b

    print(add)
    print(add.__doc__)
    add( 7, 13 )
    logfile = Path('add.log')
    assert logfile.is_file()

    # Test 03
    vspace()
    print("Test 03 : decorator 'cacheFS'")

    # @cacheFS( Path('slow_mult') )
    @cacheFS
    def slow_mult( a: int, b: int ) -> int:
        ''' Multiply a and b '''
        time.sleep(1)
        return a * b

    print(slow_mult)
    print(slow_mult.__doc__)

    @timer 
    def test03():
        ''' tests timer decorator '''
        for i in range(4):
            res = slow_mult( 111, 10*i )
            print(f"111 * {i} = {res}")

    test03() # runs in 2ms-4s depending on cache hits/misses



# pylint: disable=too-few-public-methods

# module-level docstring
__doc__='''
OS detection module
===================

Straightforward OS detection, for when platform-specific code must
be written.

I am not the original author of this class. It was downloaded from:
https://github.com/scivision/pybashutils/blob/master/pybashutils/os_detect.py

'''

from platform import system, uname

class Os:
    """
    returns class with properties:
    .cygwin   Cygwin detected
    .wsl      Windows Subsystem for Linux (WSL) detected
    .mac      Mac OS detected
    .linux    Linux detected
    .windows  Windows detected
    .bsd      BSD detected
    """

    def __init__(self):
        syst = system().lower()

        # initialize
        self.cygwin = False
        self.wsl = False
        self.mac = False
        self.linux = False
        self.windows = False
        self.bsd = False

        if 'cygwin' in syst:
            self.cygwin = True
            self.os = 'cygwin'
        elif 'darwin' in syst:
            self.mac = True
            self.os = 'mac'
        elif 'linux' in syst:
            if 'Microsoft' in uname().release:
                self.wsl = True
                self.os = 'wsl'
            else:
                self.linux = True
                self.os = 'linux'
        elif 'windows' in syst:
            self.windows = True
            self.os = 'windows'
        elif 'bsd' in syst:
            self.bsd = True
            self.os = 'bsd'

    def __str__(self):
        return self.os



# pylint: disable=broad-except

# module-level docstring
__doc__='''
CLI text-based spinning animation
=================================

An exceedingly simplistic progress animation library.
'''

import itertools
import os
import sys
from .str_utils import truncate_str

class MySpinner:
    ''' My simple spinner, one-function, easy to use and supports text
    on the left of the spinning wheel !
    
    Usage example::

        spinner = MySpinner()
        nb_items = len(items_to_process)
        for idx, item in enumerate(items_to_process):
            if idx%10==0:
                spinner.animation( text=f"Progress: {100*idx/nb_items:.1f}%" ) 
            process_item( item )
            
    '''

    def __init__( self ) -> None:
        self.last_txt_len = 0
        self.spinner = itertools.cycle(['-', '\\', '|', '/'])
        try:
            self.terminal_width = os.get_terminal_size().columns
        except Exception as e:
            print(f"MySpinner: Could not get terminal width, defaulting to 40 columns. Got error {e}")
            self.terminal_width = 40
        # print(f"MySpinner: terminal width is {self.terminal_width}")


    def animation( self, text: str = '', no_spinner: bool = False ):
        ''' update animation, with the option to print text on the left of the
        spinner character
        '''
        tmplen = (self.last_txt_len + 1)
        # erase previous message
        # why do backspaces, then whitespace, then backspaces again :
        # because backspace alone didn't consistently erase previous text
        sys.stdout.write( '\b' * tmplen + ' ' * tmplen + '\b' * tmplen ) 

        # write new message, truncated to not overflow terminal width (necessary for best results)
        spinner_symbol = '' if no_spinner else next(self.spinner)
        message = truncate_str(
            text + spinner_symbol,
            output_length=self.terminal_width-1, 
            cut_location='center'
        )
        sys.stdout.write( 
            message
        ) 
        sys.stdout.flush()
        self.last_txt_len = len( message ) - (0 if no_spinner else 1)

    def clear( self ):
        ''' erase spinner character (and accompanying text)
        '''
        self.animation( no_spinner=True )



# module-level docstring
__doc__='''
Banner module
=============

Banners are text elements used in CLI interfaces. They use is
mostly for easthetic purposes. This module provides banner 
generators and a wrapper (to bannerize functions/classes)

On formatters: two are provided
All formatters should have signature `( title: str, width: int = DEFAULT_BANNER_WIDTH, one_line: bool = False ) -> str`
(or compatible) and (for formatters in banner.py) have an entry in global dictionnary `style_switch`.
'''
 
from typing import Callable, Any, Union
import functools
import shutil
from pathlib import Path

DEFAULT_BANNER_WIDTH = min( shutil.get_terminal_size().columns, 40 )

def full_style_formatter( title: str, width: int = DEFAULT_BANNER_WIDTH, one_line: bool = False ) -> str:
    ''' Returns a "full style" banner in string representation '''
    title_len = len(title)
    d = int((width-title_len-2)/2) # number of # symbols on either side of the banner title
    second_line = "{} {}{} {}".format(
        '#'*d,
        title,
        ' ' if title_len%2==1 else '', # padding
        '#'*d
    )
    
    return second_line if one_line else "\n".join([ 
        '#' * width,
        second_line,
        '#' * width + '\n'
    ])

def lean_style_formatter( title: str, width: int = DEFAULT_BANNER_WIDTH, one_line: bool = False ) -> str:
    ''' Returns a "lean style" banner in string representation '''
    title_len = len(title)
    d = int((width-title_len-2)/2) # number of space symbols on either side of the banner title
    second_line = "#{}{}{}{}#".format(
        ' '*d,
        title,
        ' ' if title_len%2==1 else '', # padding
        ' '*d
    )
    
    return second_line if one_line else "\n".join([ 
        '#' * width,
        second_line,
        '#' * width + '\n'
    ])


style_switch = {
    'full': full_style_formatter,
    'lean': lean_style_formatter
}


def one_line_banner( title: str, style: Union[str,Callable] = 'full', width:int = DEFAULT_BANNER_WIDTH ) -> str:
    ''' Returns a string containing a one-line banner
    Available styles : 'full', 'lean'
    External formatters: If style is Callable, it is assumed to be a formatter following 
    signature norm (see module docstring)
    '''
    
    style_formatter = style if callable(style) else style_switch[style]
    banner = style_formatter( title=title, width=width, one_line=True )
    return banner + '\n'


def multi_line_banner( title: str, style: Union[str,Callable] = 'full', width:int = DEFAULT_BANNER_WIDTH ) -> str:
    ''' Returns a string containing a multi-line banner
    Available styles : 'full', 'lean'
    External formatters: If style is Callable, it is assumed to be a formatter following 
    signature norm (see module docstring)
    '''
    
    style_formatter = style if callable(style) else style_switch[style]
    banner = style_formatter( title=title, width=width )
    return banner + '\n'


def bannerize( style: Union[Callable,str], width: int = DEFAULT_BANNER_WIDTH, one_line: bool = False ) -> Callable:
    ''' Decorator. Adds a banner before function's execution. Not intended nor tested for decorating classes.
    
    When `style` is given and is a callable formatter, it will be used to produce banner, otherwise `style`
    must be a string equal to a key in `style_switch`, which defines the formatter to use.
    
    Available styles : see style_switch
    '''

    def actual_decorator( user_funtion: Callable ) -> Callable:
        ''' Sets up the banner to be displayed for user_function '''
        nonlocal style_formatter, width, one_line

        # Get name of the file containing `user_function`
        file_path = user_funtion.__getattribute__('__globals__').get('__file__')
        file_name = Path(file_path).stem
        # Setting the banner title
        _banner = style_formatter( 
            title=file_name if user_funtion.__name__=='main' else f"{file_name}.{user_funtion.__name__}",
            width=width,
            one_line=one_line
        )

        @functools.wraps(user_funtion)
        def wrapper( *args, **kwargs ) -> Any:
            ''' Prints the banner for user_function and executes it '''
            nonlocal _banner

            # Printing banner
            print( _banner )

            return user_funtion( *args, **kwargs )

        return wrapper

    if callable(style):
        # formatter passed
        style_formatter = style 
    elif isinstance(style, str):
        # `style` argument passed => validation
        style_formatter = style_switch.get( style )
        if not style_formatter:
            raise ValueError(f"bannerize: value for argument 'style' not recorgized: '{style}' not in {list(style_switch.keys())} !")
    else:
        raise ValueError("bannerize: required value for 'style' is neither a formatter (Callable) nor a string.")

    return actual_decorator



# module-level docstring
__doc__='''
MediaInfo interface
===================

An interface to software `MediaInfo <https://mediaarea.net/fr/MediaInfo>`_

Note: For it to work, MediaInfo needs to be installed on the system AND be
callable from a terminal.
'''

import logging
import re
from pathlib import Path
from typing import List, Dict, Union, Iterable

from .utils import type_assert
from .execute import execute

MEDIAINFO_POSSIBLE_RETURN_VALUES = Union[int,float,str,type(None)]

class Datapoint:
    ''' In this context, a Datapoint represents a multimedia parameter
    in a format compatible with MediaInfo's cli interface.
    
    Advanctages:
     - systematic solution
     - integrates cli<->shorthand format conversion
     - ability to combine commands
     - shifts complexity from MediaInfo's class
    '''

    SEPARATOR = '\\n'

    def __init__( self, category: str, parameter: Union[str,List[str]] ):
        self.category = category
        self.parameter = parameter

    def __str__(self):
        parameter_s = Datapoint.SEPARATOR.join( self.parameter ) \
            if isinstance( self.parameter, list ) \
            else self.parameter
        return f"--Inform={self.category};" + parameter_s

    def __repr__(self):
        return f"< Datapoint : '{str(self)}' >"

    def __parameter_shorthand(self, p: str ):
        ''' Tries to convert parameter to its equivalent shorthand.
        Helper function for `parameter_shorthand`
        '''        
        for k,v in MediaInfo.DATAPOINTS.items():
            dp = Datapoint.from_string(v)
            if dp.parameter==p:
                return k
        return p 

    @property
    def parameter_shorthand(self):
        ''' Tries to convert parameter to its equivalent shorthand '''
        return self.__parameter_shorthand( self.parameter ) \
            if isinstance( self.parameter, str ) \
            else [
                self.__parameter_shorthand( p )
                for p in self.parameter
            ]


    @classmethod
    def from_string( cls, datapoint: str ) -> 'Datapoint':
        ''' Converts datapoint from '--Inform=<category>;%parameter%'-formatted
        string to Datapoint instance
        '''
        datapoint = MediaInfo.DATAPOINTS.get( datapoint, datapoint ) # Handle shorthands
        try:
            res = re.match( r"^\-\-Inform=(.+?);(.+)$", datapoint ).groups(0)
            assert len(res)==2
            return Datapoint( *res )
        except AttributeError as e:
            raise ValueError(f"Datapoint.from_string : could not find category from datapoint string '{datapoint}'") from e


    @classmethod
    def from_strings( cls, datapoints: Iterable[str] ) -> List['Datapoint']:
        ''' Converts datapoint from '--Inform=<category>;%parameter%'-formatted
        string to Datapoint instance
        '''
        return [
            Datapoint.from_string(dp)
            for dp in datapoints
        ]


    @classmethod
    def group_by_category( cls, datapoints: Iterable['Datapoint'] ) -> Dict[str,List['Datapoint']]:
        ''' Group datapoints by category
        '''
        return {
            cat: [ 
                d
                for d in datapoints
                if d.category == cat
            ]
            for cat in { d.category for d in datapoints }
        }


    @classmethod 
    def combine( cls, datapoints: Iterable['Datapoint'] ) -> List['Datapoint']:
        ''' From an arbitrary amount of datapoints, combines them into
        Datapoints (one per category).
        '''
        categorized_datapoints = Datapoint.group_by_category(datapoints)
        
        return [
            Datapoint( 
                category=cat, 
                parameter=[
                    cat_dp.parameter
                    for cat_dp in cat_datapoints
                ]
            )
            for cat,cat_datapoints in categorized_datapoints.items()
        ]


class MediaInfo:
    ''' Interface for MediaInfo, useful for retrieving info about
    video/audio/image files.

    example : retrieving info for file 'v.mp4'::

        >>> MI = MediaInfo()                        # MediaInfo interface setup 
        >>> f = Path('v.mp4')                       # File path
        >>> MI.get_base_stats( f )                  # Returns dict with basic stats
        >>> MI.get_datapoint( f, 'V_FrameCount' )   # Returns specific stat (uses shorthand)
        >>> MI.get_datapoint( f, '--Inform=Video;%FrameCount%' )   # Equivalent to previous line

    example 2 : retrieving basic and advanced info about 'v.mp4'::

        >>> MI = MediaInfo()                        # MediaInfo interface setup 
        >>> f = Path('v.mp4')                       # File path
        >>> stats = MI.get_base_stats( f )                  # Returns dict with basic stats
        >>> stats = MI.get_datapoints( f, ['V_StreamSize','V_FrameCount','V_UniqueID'], existing_stats=stats ) # Returns specific stats (uses shorthand; update `stats`)
    
    '''

    # collection of shorthand and their MediaInfo-understandable CLI argument counterpart
    # Naming convention : '<category_name_initial>_<parameter_name>'
    DATAPOINTS = {
        'V_StreamSize' :'--Inform=Video;%StreamSize%',
        'V_FrameCount' :'--Inform=Video;%FrameCount%',
        'V_UniqueID'   :'--Inform=Video;%UniqueID%',
    }

    UNIT_FACTOR = {
        'Mb/s'    : 1_000_000,
        'kb/s'    : 1000,
        'bits'    : 1,
        'GiB'     : 1_000_000_000,
        'MiB'     : 1_000_000,
        'KiB'     : 1_000,
        'pixels'  : 1,
        'FPS'     : 1,
        'ms'      : 0.001,
        'kHz'     : 1_000,
        'channels': 1
    }

    def __init__( self, executable: str = 'MediaInfo', logLevel: int = logging.WARNING ) -> None:
        ''' `executable` : if typing 'MediaInfo' on a shell doesn't call mediainfo, specify using this argument a valid string that does.
                           It can be a variant (eg. 'mediainfo.exe') or a path (eg. '~/tmp/mediainfo/mediainfo').
        '''
        self.log = logging.getLogger('MediaInfo')
        self.log.setLevel( logLevel )
        self.executable = executable
        type_assert( executable, 'executable', str, 'MediaInfo: ' )
        self.version = None 
        self.get_MI_version()

        self.log.debug("Successfully initialized object : %s" , str(self))


    def get_MI_version( self ) -> None:
        ''' Tries to obtain callable MediaInfo version
        '''
        try:
            stdout = execute(
                command=[self.executable,'--Version']
            )['stdout'].replace('\n', '')
            self.version = re.search( r"(v[\.0-9]+)", stdout ).group(0)
        except AttributeError as e:
            raise ValueError(f"MediaInfo : could not find a version number from string '{stdout}'") from e


    def __str__(self) -> str:
        return f"< Mediainfo : executable '{self.executable}' version '{self.version}' >"


    def __repr__(self) -> str:
        return str(self)


    @classmethod
    def __try_casting_numbers( cls, s: str ) -> MEDIAINFO_POSSIBLE_RETURN_VALUES:
        ''' Simple helper function; tries to cast str -> int/float
        returns argument when fails
        '''
        
        try:
            return int(s)
        except ValueError:
            try:
                return float(s)
            except ValueError:
                return s


    @classmethod
    def __filter_data( cls, s: str ) -> MEDIAINFO_POSSIBLE_RETURN_VALUES:
        ''' MediaInfo typically returns a number. This function handles the conversion 
        process in the case this number includes a unit and a "simple"/"raw" value is
        desired. 
        
        Note that after the conversion the value is passed to `MediaInfo.__try_casting_numbers` 
        for basic int/float conversion.
        '''
        if s in ['','\r\n','\n']:
            return None
            
        for unit,mult in cls.UNIT_FACTOR.items():
            pos = s.find( unit )                
            if pos != -1:
                s_ok = re.sub( r'(\d) (\d)', r'\1\2', s[:pos] ) # remove unit and spaces between numbers
                val = MediaInfo.__try_casting_numbers( s_ok )

                if type(val) in [int,float]:
                    # successful cast
                    return val * mult
        
        return MediaInfo.__try_casting_numbers(s)


    def get_base_stats( self, media_file: Path ) -> Dict[str,Dict[str,MEDIAINFO_POSSIBLE_RETURN_VALUES]]:
        ''' Convenience function : retrieves the stats returned by MediaInfo 
        when ran without specific arguments

        WARNING: Some values returned by default by MediaInfo are imprecise 
        or rounded, typically because they are expressed in user-friendly
        units like 'Mb/s', 'GiB', etc

        Returns : <dict[<category_name:str>,<dict[<datapoint_name:str>,<MediaInfo_value>]>]>
        '''
        type_assert( media_file, 'media_file', Path, 'MediaInfo::get_base_stats: ' )
        _file = media_file.resolve()
        stdout = execute( [self.executable, str(_file)] )['stdout']

        # Preprocessing : decoding, splitting lines, regrouping datapoints
        category = 'Uncategorized'
        lines = { category : list() }
        for line in stdout.splitlines():
            
            # Discard empty lines
            if not line:
                continue
                
            # Data lines have ':'
            if ':' in line:
                lines[category].append( line.split(':') )
                continue

            # Category line (?)
            category = line
            lines[category] = list()

        # Return actual data, nicely formatted
        res = {
            category: {
                s_data[0].strip():MediaInfo.__filter_data(s_data[1].strip())
                for s_data in lines[category]
            }
            for category in lines
            if lines[category]
        }

        # cleanup
        cleanup = {
            'Video': {
                'Frame rate': r'([\.0-9]+) \([0-9/]+\) FPS'
            }
        }
        for _category,_cat_rules in cleanup.items():
            if _category not in res:
                continue

            for _datapoint, _datapoint_rule in _cat_rules.items():
                if _datapoint not in res[_category]:
                    continue
                if not isinstance( res[_category][_datapoint], str ):
                    continue

                # apply cleanup regex
                regex_res = re.search(
                    pattern=_datapoint_rule,
                    string=res[_category][_datapoint]
                )
                if regex_res:
                    res[_category][_datapoint] = self.__try_casting_numbers(
                        regex_res[1]
                    )


        return res


    def __get_datapoint( self, media_file: Path, datapoint: Datapoint ) -> MEDIAINFO_POSSIBLE_RETURN_VALUES:
        ''' Calls MediaInfo to get a datapoint value.
        `datapoint` must be in a form understandable by MediaInfo, like '--Inform=Video;%UniqueID%'
        '''

        data = execute( 
            [ 
                self.executable,
                str(datapoint),
                str(media_file)
            ]
        )['stdout']
        
        return data


    def get_datapoint( self, media_file: Path, datapoint: str ) -> dict:
        ''' Gets one datapoint's value
        For multiple datapoints, use `MediaInfo.get_datapoints`

        `datapoint` can be in a shorthand format

        Returns : <dict[<datapoint_name:str>,<MediaInfo_value>]>
        '''
        type_assert( media_file, 'media_file', Path, 'MediaInfo::get_base_stats: ' )
        _file = media_file.resolve()

        # `datapoint` may be a shorthand datapoint reference. ex : 'V_UniqueID' is shorthand for '--Inform=Video;%UniqueID%'
        dp = Datapoint.from_string( self.DATAPOINTS.get( datapoint, datapoint ) )
        return {
            dp.category : {
                dp.parameter_shorthand[2:]: MediaInfo.__filter_data(self.__get_datapoint( _file, dp ))
            }
        }


    def get_datapoints( self, media_file: Path, datapoints: List[str], existing_stats=None ) -> Dict[str,MEDIAINFO_POSSIBLE_RETURN_VALUES]:
        ''' Gets multiple datapoints with as few calls as possible.
        This is done by grouping them by category

        `datapoint` can be in a shorthand format

        `existing_stats`: to update existing stats

        Technical note : MediaInfo supports returning multiple values per call, 
        but this is limited to parameters of the same category.

        Performance note : Internal testing shows that execution time grows linearly
        with the number of calls to `MediaInfo.exec`, therefore using `MediaInfo.get_datapoints`
        is recommended over `MediaInfo.get_datapoints` when more than one datapoint is needed. (TODO: correct note)

        Returns : <dict[<datapoint_name:str>,<MediaInfo_value>]>
        '''
        type_assert( media_file, 'media_file', Path, 'MediaInfo::get_base_stats: ' )
        _file = media_file.resolve()

        # Combine datapoints
        combined_datapoints = Datapoint.combine( 
            Datapoint.from_strings(
                [
                    self.DATAPOINTS.get( dp, dp )
                    for dp in datapoints
                ]
            )
        )
        
        # Issue calls for each combined datapoint, store their result
        res = dict() if existing_stats is None else existing_stats
        for dp in combined_datapoints:
            answer = [
                MediaInfo.__filter_data( dataline )
                for dataline in self.__get_datapoint( _file, dp ).splitlines()
            ]
            nb_answer = len(answer)

            cat = dp.category

            # storing results
            if not cat in res:
                res[cat] = dict()
                
            if isinstance(dp.parameter, str) and nb_answer==1:
                # Datapoint with 1 parameter
                res[cat].update({
                    dp.parameter_shorthand[2:]:answer
                })
                continue

            if nb_answer==len(dp.parameter):
                # Datapoint with N>1 parameters
                res[cat].update({
                    p[2:]:a
                    for p,a in zip( dp.parameter_shorthand, answer )
                })
                continue

            self.log.warning("Illegal state induced by faulty values : datapoint=%s, answer='%s' for file '%s'", dp, answer, media_file)

        return res



# package-level docstring
__doc__ = '''
DRSlib - a set of utilities by DavidRodriguezSoaresCUI
======================================================

**DRSlib** is a Python package that provides a wide range of small
yet powerful set of highly-reusable functions, classes, 
decorators, etc. Its main purpose is to provide high-level
"building blocks" that accomplish simple tasks, facilitating
the writing of elaborate scripts with limited code.

License: see *LICENSE file*

Author: **DavidRodriguezSoaresCUI**

Check its `Github repository <https://github.com/DavidRodriguezSoaresCUI/DRSlib>`_ and its `Pypi project page <https://pypi.org/project/DRSlib-DavidRodriguezSoaresCUI/>`_ .
'''



# pylint: disable=too-few-public-methods, import-error, wrong-import-order, broad-except

# module-level docstring
__doc__='''
Path tools
==========

Easy to use tool to collect files matching a pattern.

Note: both class and function versions should be euivalent, both kept
just in case. Class may be usefull for repeated calls to `collect` method.

'''

from typing import Union, Iterable, List
from pathlib import Path
from os import popen
import re
import sys
import win32api
from .os_detect import Os
from .execute import execute
import logging
log = logging.getLogger( __file__ )

MAKE_FS_SAFE_PATTERN = re.compile( pattern=r'[\\/*?:"<>|]' )


class FileCollector:
    ''' Easy to use tool to collect files matching a pattern (recursive or not), using pathlib.glob. 
    Reasoning for making it a class: Making cohexist an initial check/processing on root with a recursive
    main function was not straightforward. I did it anyway, so feel free to use the function alternative. '''

    def __init__( self, root: Path ) -> None:
        assert root.is_dir()
        root.resolve()
        self.root = root
        self.log = logging.getLogger( __file__ )
        self.log.debug( "root=%s", root )

    def collect( self, pattern: Union[str,Iterable[str]] = '**/*.*' ) -> List[Path]:
        ''' Collect files matching given pattern(s) '''
        files = []
        
        if isinstance( pattern, str ):
            # 11/11/2020 BUGFIX : was collecting files in trash like a cyber racoon
            files = [
                item.resolve() 
                for item in self.root.glob( pattern )
                if item.is_file() and (not '$RECYCLE.BIN' in item.parts)
            ]

            self.log.debug( "\t'%s': Found %s files in %s", pattern, len(files), self.root )
        elif isinstance( pattern, Iterable ):
            patterns = pattern
            assert 0 < len(patterns)
            for p in patterns:
                files.extend( self.collect(p) )
        else:
            raise ValueError(f"FileCollector: 'pattern' ({pattern}) must be an Iterable or a string, but is a {type(pattern)}")

        return files
        

def file_collector( root: Path, pattern: Union[str,Iterable[str]] = '**/*.*' ) -> List[Path]:
    ''' Easy to use tool to collect files matching a pattern (recursive or not), using pathlib.glob.
    Collect files matching given pattern(s) '''
    assert root.is_dir()
    root.resolve()
    log.debug( "root=%s", root )

    def collect( _pattern: str ) -> List[Path]:
        # 11/11/2020 BUGFIX : was collecting files in trash like a cyber racoon
        _files = [
            item.resolve() 
            for item in root.glob( pattern )
            if item.is_file() and (not '$RECYCLE.BIN' in item.parts)
        ]
        log.debug( "\t'%s': Found %s files in %s", pattern, len(_files), root )
        return _files

    files = []
    if isinstance( pattern, str ):
        files = collect( pattern )
    elif isinstance( pattern, Iterable ):
        patterns = pattern
        assert 0 < len(patterns)
        for p in patterns:
            files.extend( collect(p) )
    else:
        raise ValueError(f"FileCollector: 'pattern' ({pattern}) must be an Iterable or a string, but is a {type(pattern)}")

    return files


def make_FS_safe( s: str ) -> str:
    ''' File Systems don't accept all characters on file/directory names.

    Return s with illegal characters stripped

    Note: OS/FS agnostic, applies a simple filter on characters: ``\\, /, *, ?, :, ", <, >, |``
    '''
    return re.sub(
        pattern=MAKE_FS_SAFE_PATTERN,
        repl="",
        string=s
    )

def find_available_path( root: Path, base_name, file: bool ) -> Path:
    ''' Returns a path to a file/directory that DOESN'T already exist.
    The file/dir the user wishes to make a path for is referred as X.

    `root`: where X must be created. Can be a list of path parts
    `base_name`: the base name for X. May be completed with '(index)' if name already exists.
    `file`: True if X is a file, False if it is a directory
    '''
    # Helper function: makes suffixes for already existing files/directories
    def suffixes():
        yield ''
        idx=0
        while True:
            idx+=1
            yield f" ({idx})"
    
    # Iterate over candidate paths until an unused one is found
    safe_base_name = make_FS_safe( base_name )
    if file:
        # name formatting has to keep the extension at the end of the name !
        ext_idx = safe_base_name.rfind('.')
        assert ext_idx!=-1
        f_name, f_ext = safe_base_name[:ext_idx], safe_base_name[ext_idx:]
        for suffix in suffixes():
            _object = root / ( f_name + suffix + f_ext )
            if not _object.is_file():
                return _object
    else:
        for suffix in suffixes():
            _object = root / ( safe_base_name + suffix )
            if not _object.is_dir():
                return _object


def make_valid_path( 
    root: Union[Path, List],
    base_name: str,
    file: bool = True,
    create: bool = False ) -> Path:
    ''' Returns a path to a file/directory that DOESN'T already exist.
    The file/dir the user wishes to make a path for is referred as X.

    `root`: where X must be created. Can be a list of path parts
    `base_name`: the base name for X. May be completed with '(index)' if name already exists.
    `file`: True if X is a file, False if it is a directory
    `create`: True instantiates X (empty file or dir), False doesn't

    Build upon `find_available_path`, adding:
    - root path construction
    - root mkdir
    - ability to initialize returned file/dir
    '''

    # make root path
    if isinstance(root, List):
        if isinstance(root[0], str):
            _root = Path(make_FS_safe(root[0]))
        elif isinstance(root[0], Path):
            _root = root[0]
        else:
            raise TypeError(f"root[0]={root[0]} is of unexpected type {type(root[0])}, not str or Path !")
        
        for path_part in root[1:]:
            assert isinstance(path_part, str), f"path part in root '{path_part}' is of unexpected type {type(path_part)}, not str !"
            safe_path_part = make_FS_safe(path_part)
            assert safe_path_part
            _root = _root / safe_path_part
    elif isinstance(root, Path):
        _root = root
    else:
        raise TypeError(f"root={root} is of unexpected type {type(root)}, not str or Path !")
            
    # make root directory
    ensure_dir_exists( _root )

    # Find valid path
    valid_path = find_available_path(
        _root,
        base_name,
        file
    )

    # Optionally create file/dir
    if create:
        if file:
            valid_path.touch()
        else:
            valid_path.mkdir()

    return valid_path


def ensure_dir_exists( folder: Path ) -> None:
    ''' Tests whether `folder` exists, creates it (and its whole path) if it doesn't.
    '''
    if folder.is_file():
        raise ValueError(f"Given path '{folder}' is a file !")
    if not folder.is_dir():
        folder.mkdir( parents=True )


def folder_get_file_count( _root: Path, use_fallback: bool = False ) -> int:
    ''' Uses built-in platform-specific ways to recursively count the number of files in a given directory.
    Reason for using CLI calls to platform-specific external tools : they typically offer superior performance (because optimised)

    `use_fallback` : if True, use Path.glob instead of platform-specific CLI calls (mainly for testing puposes)
    '''
    _root = _root.resolve()

    def fallback() -> int:
        return sum( 1 for x in _root.glob('**/*') if x.is_file() )

    if use_fallback:
        return fallback()

    current_os = Os()
    command = None
    if current_os.windows:
        # Windows CMD
        log.debug("Crawler from '%s'", _root)
        command = f"dir \"{_root}\" /A:-D /B /S | find \".\" /C"
    elif current_os.wsl or current_os.linux:
        # Linux
        log.debug("Crawler from '%s'", _root)
        command = f"find \"{_root}\" -type f|wc -l"
    else:
        log.warning( "OS not recognised or has no specific command set (%s); fallback method used.", current_os )
        return fallback()

    
    return int(popen( command ).read().strip())


def folder_get_subdirs( root_dir: Path ) -> List[Path]:
    ''' Return a list of first level subdirectories '''
    assert root_dir.is_dir()
    return [ 
        item
        for item in root_dir.resolve().iterdir()
        if item.is_dir() and (not '$RECYCLE.BIN' in item.parts)
    ]


def windows_list_logical_drives() -> List[Path]:
    ''' Uses windows-specific methods to retrieve a list of logical drives.

    Both methods have been developped and tested to give equivalent output and be interchangeable
    '''
    
    def method1():
        ''' uses a windows shell command to list drives '''
        try:

            def filter_drives( l ):
                for item in l:
                    if not item:
                        continue
                    try:
                        yield Path(item).resolve()
                    except Exception:
                        continue

            drives = list( 
                filter_drives( win32api.GetLogicalDriveStrings().split('\x00') ) 
            )
            return drives
        except ImportError as e:
            print(f"windows_list_logical_drives: ImportError raised, so you may have tried using this function on a non-windows system, or the `pypiwin32` package is missing : {e}")
            exit(1)
    
    def method2():
        ''' uses a windows shell command to list drives '''
        command = [ 'wmic', 'logicaldisk', 'get', 'name' ]
        stdout = execute( command )['stdout']

        def return_cleaned( l ):
            for item in l:
                if len(item) < 2:
                    continue
                if item[0].isupper() and item[1] == ':':
                    try:
                        # Bugfix : the '<driveletter>:' format was resolving to CWD when driveletter==CWD's driveletter. 
                        # This seems to be an expected Windows behavior. Fix: switch to '<driveletter>:\\' format, whis is more appropriate.
                        yield Path(item[:2]+'\\').resolve()
                    except Exception:
                        continue
        
        drives = list( return_cleaned(stdout.splitlines()) )
        return drives
    
    # Test
    # assert all( [ x.samefile(y) for x,y in zip(method1(),method2()) ] )
    
    try:
        return method1()
    except Exception:
        try:
            return method2()
        except Exception as e:
            print(f"windows_list_logical_drives: something went wrong : {e}")
            sys.exit(1)



# module-level docstring
__doc__='''
Hash-related tools
==================

Useful tools for easily hashing files.
'''

import hashlib
from pathlib import Path
import logging
from typing import Union, Optional, Dict, Iterable

from .utils import pickle_this, unpickle_this
from .spinner import MySpinner
from .os_detect import Os
from .path_tools import ensure_dir_exists, file_collector

BLOCKSIZE = 65_536 # 2 ** 16
log = logging.getLogger( __file__ )
spinner = MySpinner()
current_os = Os()

def file_hash( file: Path ) -> str:
    ''' Returns the md5 hash of the corresponding file
    Reads the file in 2^16 bytes (~65Ko) chunks
    '''

    hasher = hashlib.md5()
    with file.open(mode='rb') as f:
        chunk = f.read(BLOCKSIZE)
        while 0 < len(chunk):
            hasher.update(chunk)
            chunk = f.read(BLOCKSIZE)

    hash_s = hasher.hexdigest()

    log.debug( "File %s has md5 hash %s.", file, hash_s )
    return hash_s


def partial_MD5( file: Path ) -> str:
    ''' Reads up to 10MB of the given file and returns MD5 (partial) checksum. Borrowing code from : https://stackoverflow.com/a/1131238
    This is achieved by reading 10 times ~1MB, thus reducing RAM usage.
    '''
    hasher = hashlib.md5()
    with file.open(mode='rb') as f:
        for _ in range(10):
            chunk = f.read(1_048_576)
            if not chunk:
                break
            hasher.update(chunk)
            
    return hasher.hexdigest()


def path_to_id( _path: Path, limit_id_len: bool = True ) -> str:
    ''' Returns a string identifier for any given file/directory path.

    `limit_id_len`: If True, keeps id length below 190 characters
    '''
    path = _path.resolve()

    anchor_len = 3 if current_os.wsl else len(path.anchor)

    _id = str(path)[anchor_len:]
    if limit_id_len and 190 < len(_id):
        _id = _id[:160] + ' [...] ' + _id[-20:]

    return _id


def directory_hash( directory: Path, pattern: Union[str,Iterable[str]] = '*.*', old_hashes: Optional[dict] = None ) -> Dict[str,str]:
    ''' Returns a dictionnary of md5 hash of all the files in the directory
    corresponding to path 'directory'.
    
    Returns: Dict[ <file_name:str>, <file_hash:str> ]
    '''

    files = file_collector(
        root = directory,
        pattern=pattern
    )

    hashes = dict()
    for file in files:
        key = file.name
        if key in old_hashes:
            log.debug('Cache hit')
            hashes[key] = old_hashes[key]
            continue
            
        hashes[key] = file_hash( file )        

    return hashes


def tree_hash( root: Path, hash_location: Path, pattern: Union[str,Iterable[str]] = '*.*', fastload: bool = False ) -> None:
    ''' Explores the directory structure recursively from 'root', computing their hash.
    For each directory explored, saves hashes found to a file.

    `fastload`: If True and the corresponding hash file is found, hashes are not re-checked. This
    is used for performance reasons. Do not use if changes are likely.

    '''

    root = root.resolve()
    hash_location = hash_location.resolve()
    assert root.is_dir()
    ensure_dir_exists( hash_location )

    subdirectories = [x for x in root.iterdir() if x.is_dir() and x!=hash_location]
    for sub_dir in subdirectories:
        savefile = hash_location / path_to_id( sub_dir )

        if not (fastload and savefile.is_file()):
            old_hashes = unpickle_this( savefile )
            spinner.animation()
            hashes = directory_hash( 
                directory=sub_dir,
                pattern=pattern,
                old_hashes=old_hashes
            )
            data = dict()
            data['folder'] = sub_dir
            data['hashes'] = hashes
            
            pickle_this(data, savefile)
        else:
            spinner.animation(text='cache hit')
        
        # Recursion
        tree_hash(
            root=sub_dir,
            hash_location=hash_location,
            pattern=pattern,
            fastload=fastload
        )




# module-level docstring
__doc__='''
General Utils module
====================

A collection of tools that do not fit in other modules.
'''

import pickle
from pathlib import Path
from typing import Any, Union, Iterable
import logging
log = logging.getLogger( __file__ )

def pickle_this( data: Any, save_file: Path ) -> None:
    ''' Stores `data` to a file
    '''
    with save_file.open(mode="wb") as fp:   #Pickling
        pickle.dump(data, fp)
        log.debug( "Successfully stored data to file %s.", save_file )
        

def unpickle_this( save_file: Path ) -> Any:
    ''' Reads data from a file that was created by `pickle_this`
    '''
    if not save_file.is_file():
        log.debug( "Could not find file %s.", save_file )
        return None
    with save_file.open(mode="rb") as fp:   # Unpickling
        log.debug( "Successfully retrieved data from file %s.", save_file )
        return pickle.load(fp)


def is_iterable( a: Any ) -> bool:
    ''' Checks if object is iterable, thus having attribute `__iter__`.
    '''
    return hasattr(a, "__iter__")


def type_assert( v: object, v_name: str, expected_type: Union[type, Iterable[type]], prefix: str, accept_none: bool = False ) -> None:
    ''' Makes sure a variable is of given type(s).

    Raises AssertionError if `v` has type not covered by `expected_type`.

    `prefix` : A string prefix for the message the raised error may display, useful for knowing which
               part of the code generated the error. eg: 'MyClass.method2'

    `accept_none` : if True, accepts a `None` value for `v` no matter the expected type(s). 
                    Alternatively, include `type(None)` to the expected type list.
    '''
    # Case 1 : accept None value
    if accept_none is True and v is None:
        return

    error_txt = f"{prefix}: expected received '{v_name}' to be a {expected_type} object, received '{type(v)}'"
    # Case 2 : expected_type is a list/set => recursion
    if is_iterable(expected_type):
        for e_t in expected_type:
            try:
                type_assert( v, v_name, e_t, prefix, accept_none )
                return
            except AssertionError:
                continue
        raise AssertionError( error_txt )
    
    # Default case : check type
    assert isinstance( v, expected_type ), error_txt



# module-level docstring
__doc__='''
Interval representation implementation
======================================

A simple implementation for generic intervals.
'''


from typing import NewType, Union, Optional, Tuple

Numerical = NewType( 'Numerical', Union[float,int] )

class Interval:
    ''' Represents an interval [a,b] of real (numerical) values
    Note : the interval [a,b] must be strictly positive (a<b)
    '''

    class InvalidInterval(Exception):
        ''' Trivial exception for invalid interval '''
        
    # Needed for visibility
    InvalidInterval=InvalidInterval

    def __init__( self, a: Numerical, b: Numerical, boundary: Optional['Interval'] = None ) -> None:
        assert a<b, f"Interval.__init__: invalid arguments : a={a}, b={b} do not respect rule a<b"
        self.a = a
        self.b = b
        if isinstance(boundary, Interval):
            self.a, self.b = self.intersection( boundary )      

    @property
    def len(self) -> Numerical:
        ''' Property. Represents the length of the interval '''
        return self.b - self.a

    def __str__( self ):
        return f"<Interval {self.a:.2f}-{self.b:.2f} [{self.len:.2f}]>"

    def intersection( self, other: 'Interval', return_obj: bool = False, raise_exception: bool = True ) -> Union['Interval',Tuple[Numerical,Numerical], None]:
        ''' Computes the intersection between two Interval objects.
        Let [x,y] be the valid strictly interval representing the intersection (only exists on
        strictly overlapping intervals).
        
        Returns : either an Interval object representing the [x,y] interval (if `return_obj`),
        or the (x,y) tuple (if not `return_obj`), or nothing if [x,y] doesn't exist 
        (see `raise_exception` below).
        
        `raise_exception` : allows to disable raising exception, returning None instead.
        '''
        assert isinstance( other, Interval )
        a = self.a if other.a < self.a else other.a
        b = self.b if self.b < other.b else other.b
        if b <= a and raise_exception:
            # non-strictly overlapping intervals 
            raise self.InvalidInterval()
        else:
            # valid intersection exists 
            return Interval(a,b) if return_obj else (a, b)

    @classmethod
    def from_length( cls, length: Numerical, start: Optional[Numerical] = None, end: Optional[Numerical] = None ) -> 'Interval':
        ''' Creates an Interval from 2 (or more) of an interval's properties : start point, end point, length '''
        assert 0 < length
        if start:
            if end:
                assert abs((end - start) - length) < (end - start) * 0.01, "Interval.from_length: argument mismatch"
            else:
                end = start + length
        elif end:
            start = end - length
        return Interval( start, end )
        


# pylint: disable=too-many-return-statements

# module-level docstring
__doc__='''
String-focused utils
====================

Simple utils that accomplish one task well, specifically string operations.

'''
 
import re

def ensure_double_quotes( s: str ) -> str:
    ''' Ensures s is double quoted in a particular way :
     - returned string must begin and end with a double quote character `"`
     - any double quote that isn't at the beginning or end must be escaped `\\"`

    Examples :
     - `` (empty string) -> `""`
     - `.` -> `"."`
     - `hello` -> `"hello"`
     - `"hello` -> `"hello"`
     - `hello"` -> `"hello"`
     - `"hello"` -> `"hello"` (unchanged)
     - `"he"llo` -> `"he\\"llo"`
     - `"he\\"llo"` -> `"he\\"llo"` (unchanged)

    '''

    if not s:
        return '""'
    if len(s)==1:
        if s[0]=='"':
            return '"\\""'
        return '"' + s + '"'
    # else
    if s[0] != '"':
        return ensure_double_quotes( '"' + s )
    if s[-1] != '"':
        return ensure_double_quotes( s + '"' )

    # past this point, 2 <= len(s) and there are double quotes at the beginning and end of s.
    
    double_quote_locations = set( m.start() for m in re.finditer(r'"', s) )
    escaped_double_quote_locations = set( m.start()+1 for m in re.finditer(r'\\"', s) )
    double_quote_locations.difference_update( escaped_double_quote_locations )

    # s lacks any double quotes
    if not double_quote_locations:
        return '"' + s + '"'

    # detection of double quotes to escape
    proper_double_quote_location = set( [0, len(s)-1] )
    misplaced_double_quotes = set(double_quote_locations).difference( proper_double_quote_location )
    # print(f"proper_double_quote_location={proper_double_quote_location}")
    # print(f"double_quote_locations={double_quote_locations}")
    # print(f"misplaced_double_quotes={misplaced_double_quotes}")
    
    # s is already properly double quoted
    if not misplaced_double_quotes:
        return s

    # escaping double quotes plus recursion
    s_characters = list(s)
    misplaced_DQ_idx = misplaced_double_quotes.pop()
    _s = ''.join( s_characters[:misplaced_DQ_idx] + [ '\\', '"' ] + s_characters[misplaced_DQ_idx+1:] )
    return ensure_double_quotes( _s )


def truncate_str( s: str, output_length: int, cut_location: str = 'center' ) -> str:
    ''' Truncates string to a new length
    
    `cut_location` : (default:'center') in ['left','center','right']
    ''' 
    s_len = len(s)
    if s_len <= output_length:
        return s 
    if output_length <= 7:
        print(f"Warning: used truncate_str with 'output_length'={output_length}<=7, which is too low and would probably result in undesired results, so 's' was returned unchanged !")
        return s

    #len_diff = s_len-output_length
    if cut_location=='left':
        return f"{ s[:output_length-6] } [...]"
    if cut_location=='center':
        offset = (output_length // 2) - 3
        return f"{ s[:offset] } [...] { s[-(offset - (1 if output_length%2==0 else 0)):] }"

    if cut_location=='right':
        return f"[...] { s[-(output_length-6):] }"
    # else
    raise ValueError(f"truncate_str: given parameter 'cut_location'={cut_location} is not in ['left','center','right'] !")


if __name__=='__main__':

    # ensure_double_quotes tests

    # tests = [
    #     '',
    #     '.',
    #     '"',
    #     '"hello',
    #     'hello"',
    #     '"hello"',
    #     '"he"llo"',
    #     'he\\"llo'
    # ]

    # for test_s in tests:
    #     res = ensure_double_quotes(test_s)
    #     print(f"test_s={test_s}={list(test_s)} -> {list(res)} = {res}")


    # truncate_str tests

    tests = [
        'h'* 20,
        'h' * 21
    ]
    for new_len in [10,15]:
        for cut in ['left','center','right']:
            for test_s in tests:
                res = truncate_str( test_s, new_len, cut)
                print(f"new_len={new_len}, cut={cut} : test_s='{test_s}' (len={len(test_s)}) -> '{res}' (len={len(res)})")



# pylint: disable=broad-except, eval-used

# module-level docstring
__doc__='''
FileSystemDataBase
==================

A collection of tools for having a cached representation of
a file system.
'''

from pathlib import Path
import re
import json
from typing import Callable, Union, List, Dict, Iterable
import logging
from .decorators import timer
from .path_tools import ensure_dir_exists, make_FS_safe, folder_get_file_count
from .utils import pickle_this, unpickle_this
log = logging.getLogger( __file__ )


def FSindex( root: Path, condition: Callable = lambda x: True ) -> Dict[str,dict]:
    ''' Returns a dictionnary such as
    - Contains <root_path:str> as only key
    - Recursively represents contained files and subdirectories, with each their subdirectories, etc
    '''
    assert root.is_dir()
    _root = root.resolve()

    def recursive_collection( _dir: Path ):
        def try_recursion( location: Path ):
            try:
                return recursive_collection( location ) if location.is_dir() else None
            except Exception as e:
                print(f"FSindex: something went wrong at '{root}'. Error:\n{e}")
                return None

        return {
            str(child.name): try_recursion( child )
            for child in _dir.iterdir()
            if condition( child )
        }

    root_s = root.as_posix()
    if root_s[-1] == '/':
        root_s = root_s[:-1]

    return {
        root_s: recursive_collection( root )
    }


class CachedFS:
    ''' Caching a filesystem tree can be useful for applications
    with frequent file system lookups.
    
    Features :
     - Possiblity to backup to/load from JSON file
     - cache directories, files, or both
    '''

    def __init__( self, root: Path, directories: bool = True, files: bool = True, backup_fs: dict = None ) -> None:

        if backup_fs:
            self.root = Path( backup_fs['root'] )
            assert root.samefile(self.root), f"ERROR: given root path '{root}' is different from backup root path '{self.root}' !"
            assert self.root.is_dir(), f"It seems root='{self.root}' is no longer a valid directory !"
            self.filter_text = backup_fs['filter']
            self.filter = eval( backup_fs['filter'] )
            self.fs = backup_fs['fs']

        else:
            assert files or directories, "CachedFS cannot be instantiated with directories=files=False."
            assert root.is_dir(), f"root='{root}' is not a valid directory"
            self.root = root.resolve()
            condition = ' or '.join( [
                x
                for x in [
                    'x.is_dir()' if directories else None,
                    'x.is_file()' if files else None
                ]
                if x is not None 
            ] )
            self.filter_text = f'lambda x: {condition}'
            self.filter = eval( self.filter_text )
            self.update()
        
        
    @timer
    def update( self ) -> None:
        ''' Updates internal DB '''
        self.fs = FSindex( self.root, self.filter )


    def __contains__( self, pattern: str ) -> bool:
        ''' Implements `<pattern:str> in <_:CachedFS>` operation
        '''
        _pattern = re.compile(pattern, flags=re.IGNORECASE )

        return 0 < len( self.search( 
            search_for=_pattern,
            stop_at_first=True
        ) )


    @timer
    def search( self, search_for: Union[str,re.Pattern,Callable], stop_at_first: bool = False ) -> List[str]:
        ''' Performs a search on internal FileSystem representation.

        `search_for`: Match criterion. Can be a string (matches file/directory name; case insensitive),
        a re.Pattern (matches with re.Pattern.match()) or a callable (must return boolean values; match
        on True).

        `stop_at_first`: returns at most one matching item.
        '''

        if callable(search_for):
            _search_for = search_for
        if isinstance(search_for, re.Pattern):
            _search_for = lambda x: bool(search_for.match(x))
        if isinstance(search_for, str):
            search_for_lower = search_for.lower()
            _search_for = lambda x: search_for_lower in x.lower()

        combine_paths = lambda x,y: f"{x}/{y}" if x else y

        def search_recursive( FSDB: dict, path_prefix: str = '' ):
            matches = list()
            #print(f"search_recursive: from '{path_prefix}'")
            try:
                for item, children in FSDB.items():
                    if _search_for( item ):
                        #print(f"Positive: '{item}'")
                        matches.append( combine_paths( path_prefix, item ) )
                        if stop_at_first:
                            return matches
                    if children:
                        matches.extend( 
                            search_recursive( children, combine_paths( path_prefix, item ) )
                        )
                    if stop_at_first and matches:
                        return matches
            except Exception as e:
                print(f"search_recursive: something went wrong at '{path_prefix}'. Error:\n{e}")
                raise
            
            return matches
        
        return search_recursive( self.fs )


    def as_json( self ) -> str:
        ''' Dumps CachedFS as json-formatted string '''
        data = {
            'root': str(self.root),
            'fs': self.fs,
            'filter': self.filter_text
        }
        return json.dumps( data, indent=2 )

    
    def backup_to_file( self, backup_file: Path ) -> None:
        ''' Dumps CachedFS to json-formatted file '''
        assert backup_file.suffix.lower() == '.json'
        backup_file.write_text( self.as_json(), encoding='utf8' )


    @classmethod
    def from_file( cls, backup_file: Path, root: Path ):
        ''' Loads CachedFS from json-formatted string produced by CachedFS.backup_to_file() function.
        Note: `root` is required to verify the intended root matches root in cache file.
        '''
        assert backup_file.suffix.lower() == '.json'
        backup = json.loads( backup_file.read_text( encoding='utf8' ) )
        return CachedFS( root=root, backup_fs=backup )
    

def get_snapshot_file( snapshot_folder: Path, snapshot_filename: str ) -> Path:
    ''' Returns a snapshot file path, given snapshot folder path and a filename 
    (which passes through a sanitization process).
    '''
    
    ensure_dir_exists( snapshot_folder )
    return snapshot_folder / make_FS_safe( snapshot_filename + '.snapshot' )


def get_folder_snapshot_h( _root: Path, extentions: Iterable[str], snapshot_folder: Path, recursive: bool = True, simplified: bool = False, rec: bool = False ) -> dict:
    ''' Recursive helper function to `get_folder_snapshot`.
    For more information see its docstring.
    '''

    # '*.txt' -> 'txt' extension conversion
    extentions = [ e.replace('*.','.') for e in extentions ]

    # Check for cached results
    cache_file_name = str(_root) + ('.S' if simplified else '.C')
    cache_file = get_snapshot_file( snapshot_folder, cache_file_name )
    try:
        res = unpickle_this( cache_file )
    except Exception:
        res = None
    if res:
        if simplified:
            log.info( "get_folder_snapshot: loaded from cache" )
            return res
        # Check for discrepancy in file number, as indicator something changed and snapshot should be rebuilt
        nbfiles1, nbfiles2 = res['__nbfiles__'], folder_get_file_count( _root )
        if nbfiles1==nbfiles2:
            log.info( "get_folder_snapshot: loaded from cache" )
            return res
        log.debug("get_folder_snapshot: number of files changes (%s != %s) -> updating snapshot", nbfiles1, nbfiles2)

    # Build from scratch
    log.info("get_folder_snapshot: from '%s' ..", _root)
    content = dict()

    if simplified: # "simplified" version
        for item in _root.iterdir():
            if item.is_file():
                (name, ext) = (item.stem, item.suffix)
                if ext not in extentions:
                    continue
                content[name] = item.resolve()
            elif item.is_dir():
                if not recursive:
                    continue
                content_rec = get_folder_snapshot_h( item, extentions, snapshot_folder, recursive, simplified, rec=True )
                content.update( content_rec )
            else:
                raise ValueError("get_folder_snapshot::Not file or folder : {}".format(item))
                
    else: # "complex" version
        file_count = 0
        for item in _root.iterdir():
            if item.is_file():
                (name, ext) = (item.stem, item.suffix)
                if ext not in extentions:
                    continue
                content[name] = item.resolve()
                file_count += 1
            elif item.is_dir():
                if not recursive:
                    continue
                content_rec = get_folder_snapshot_h( item, extentions, snapshot_folder, recursive, simplified, rec=True )
                content[item.name] = content_rec
                file_count += content_rec['__nbfiles__']
            else:
                raise ValueError("get_folder_snapshot::Not file or folder : {}".format(item))

        # Add file count
        content['__nbfiles__'] = file_count
        log.debug("get_folder_snapshot: from '%s' : found %s files !", _root, file_count)

    # Caching for later use
    if not (simplified and rec):
        pickle_this( content, cache_file )

    return content


def get_folder_snapshot( _root: Path, extentions: Iterable[str], snapshot_folder: Path, recursive: bool = True, simplified: bool = False ) -> dict:
    ''' Recursively builds a dictionnary representation of a directory tree. Only listed file extensions are considered.

    `extensions` : iterable of ``'*.ext'`` or ``'.ext'``

    Rules ("complex") :
     - The returned dictionnary has one entry : 'root'
     - An entry E can have one of three values according to  :
       > E is a directory : its value is a dictionnary with name of contained items as keys
       > E is a file : its value is its path (string format)
     - Each directory has one special entry '__nbfiles__', the recursive file count

    ex : let the following tree::

        <root>
            --A
                --T
                    -->file8.png
                -->file7.txt
                -->file4.jpg
            --C
                -->file5.git
            --G
            -->file3.log

    will have representation ::

        {
            '<root>': {
                'A': {
                    'T': {
                        'file8': '<root>/A/T/file8.png',
                        '__nbfiles__': 1
                    },
                    'file7': '<root>/A/file7.txt',
                    'file4': '<root>/A/file4.jpg',
                    '__nbfiles__': 3
                },
                'C': {
                    'file5': '<root>/A/file5.git',
                    '__nbfiles__': 1
                },
                'G': None,
                'file3' : '<root>/file3.log',
                '__nbfiles__': 5
            }
        }

    Rules ("simplified"):
     - Returns a dictionnary without nesting
     - All entries are files (with matching extension) with format : <filename:str> : <absolute file path:Path>

    Representation for previous tree::

        {
            'file3': '<root>/file3.log',
            'file4': '<root>/A/file4.jpg',
            'file5': '<root>/A/file5.git',
            'file7': '<root>/A/file7.txt',
            'file8': '<root>/A/T/file8.png'
        }
    '''
    ensure_dir_exists( snapshot_folder )
    return get_folder_snapshot_h( _root, extentions, snapshot_folder, recursive, simplified )



# pylint: disable=eval-used, broad-except

# module-level docstring
__doc__='''
Command line user interface
===========================

Implements convenience function for CLI user interaction.
Useful when you need to ask the user what to do, or select one
of many options.
'''

from typing import Iterable, Union, Any, Callable, Dict, Optional
from pathlib import Path
import sys
import os
from .banner import one_line_banner
from .os_detect import Os
from .path_tools import folder_get_subdirs, windows_list_logical_drives, make_FS_safe

KBI_msg = "A KEYBOARDINTERRUPT WAS RAISED. THE PROGRAM WILL EXIT NOW."

def __input_KBI( message: str, exit_on_KBI: bool = True ) -> str:
    ''' Handles `KeyboardInterrupts` on `input` calls, used by other more complex functions.
    `exit_on_KBI`: If True, user can exit the program. If False, handling
    of KeyboardInterrupt is delegated to calling code.
    '''
    if exit_on_KBI:
        try:
            return input( message )
        except KeyboardInterrupt:
            print( KBI_msg )
            end_of_program()
    return input( message )


def pause() -> None:
    ''' Implements a 'pause' feature. Press ENTER to continue. If 'Ctrl+C' is pressed, 
    it exits the program '''

    __input_KBI( "Press the <ENTER> key to continue..." )


def end_of_program( exit_code: int = 0, halt: bool = False ) -> None:
    ''' Standardized way of ending programs '''
    print("\nEND OF PROGRAM\n")
    if halt:
        pause()
    sys.exit(exit_code)



def yes_or_no( message: str, retry: bool = True, default: bool = None, exit_on_KBI: bool = True ) -> bool:
    ''' User input fiels optimized for obtaining y/n answer.
    Loops user input until `KeyboardInterrupt` is raised (hard exit) or user types correct answer.
    `default` is returned if user input is inconclusive and `retry` is False.

    eg::

        >>> continue = yes_or_no( "Continue ?", retry=False, defalut=False )
        'Continue ? (y/n): '
        >>> confirm_date = yes_or_no( "Is the date correct ?" )
        'Is the date correct ? (y/n): '
        
    '''
    if not retry:
        assert default is not None and isinstance( default, bool )

    while True:
        default_txt = ' [default:y]' if default is True else ( ' [default:n]' if default is False else '' )
        raw_input = __input_KBI( 
            f"{message} (y/n){default_txt}: ",
            exit_on_KBI=exit_on_KBI
        ).lower()

        if raw_input=='' and default is not None:
            return default
        if raw_input[0] == 'y':
            return True
        if raw_input[0] == 'n':
            return False
        if retry:
            print("Invalid answer, try again")
            continue
        return default


def user_input( prompt: str, accepted: Union[Iterable[Union[str,int]],Callable], default: Any = None ) -> str:
    ''' Asks user for input, with restrictions on accpetable values.
    `prompt`: appropriate text asking the user for input. Should be straightforward and informative about the kind of data that is asked
    `accepted`: either a function testing if the user input is acceptable, or an iterable containing all acceptable values
    `default`: When given, if the user input is not acceptes, default is returned. When abscent, the user will be prompted again until either
    an accepted value is entered or a KeyboardInterrupt is raised.
    Note: this is only designed to retrieve values of the following types: str, int, float
    '''
    if default is not None:
        prompt += f"[default:{default}] "

    acceptable_UI = lambda ui: (callable(accepted) and accepted(ui)) or (ui in accepted)
        
    while True:
        # main loop: ask user until an acceptable input is received, or a KeyboradInterrupt ends the program
        _user_input = __input_KBI( prompt )
        
        # case: raw user input is accepted
        if acceptable_UI( _user_input ):
            return _user_input
        
        # case: processed user input is accepted
        variations = [ 'int(_user_input)', 'float(_user_input)', '_user_input.lower()' ]
        for variation in variations:
            try:
                __user_input = eval( variation )
                if _user_input( __user_input ):
                    return __user_input
            except Exception:
                pass
        
        # case: user input is not accepted AND there is a default
        if default is not None:
            return default
        
        # case: user input is not accepted AND there is no default => notify user, ask again
        print(f"Input {_user_input} is not a valid input. %s", f"Please choose one of : {accepted}" if not callable(accepted) else "" )


def choose_from_list( choices: list, default: int ) -> Any:
    ''' Prints then asks the user to choose an item from a list
    `default`
    '''
    # Print choices
    print( "Choices:\n  " + '\n  '.join([
        f"[{idx}] {choice}"
        for idx, choice in enumerate(choices)
    ]) + '\n' )

    # Get user selection
    idx = user_input( "Selection : ", accepted=list(range(len(choices))), default=default )

    # Return choice
    return choices[idx]


def select_action( choices: Dict[str,Callable], no_banner: bool = False, default: str = None, execute: bool = False ) -> Optional[Callable]:
    ''' Asks the user to choose an action amongst a list of labeled actions. Returns a callable
    corresponding to the chosen action if `execute` is False, executes it otherwise.
    
    Example of `choices`::

        choices = {
            'q': {
                'explanation': 'quit the program',
                'action': <function that exits the program:Callable>
            },
            ...
        }
    ''' 

    # Print banner
    if not no_banner:
        banner = one_line_banner( "Selection menu" )
        print(banner)

    accepted_inputs = choices.keys()

    # Print choices
    choice_formatter = lambda choice: choice + ( f" - {choices[choice]['explanation']}" if 'explanation' in choices[choice] else '' )
    print( "Choices:\n  " + '\n  '.join([
        choice_formatter( choice )
        for choice in accepted_inputs
    ]) + '\n' )
    
    # Get user choice
    _user_input = user_input( "Selection : ", accepted=accepted_inputs, default=default )

    # Return or execute corresponding callable
    if execute:
        choices[_user_input]['action']()
    else:
        return choices[_user_input]['action']


def cli_explorer( root_dir: Path = None, allow_mkdir: bool = True ) -> Path:
    ''' Allows for the user to explore directories to select one.
    '''

    sub_dirs = folder_get_subdirs( root_dir ) if root_dir else windows_list_logical_drives()
    cwd = root_dir
    
    def get_parent():
        try:
            return cwd.parent
        except Exception:
            return None

    while True:
        print(f"cwd : {cwd}")

        # Craft selection list
        selection_list = [ d.name if 0<len(d.name) else str(d) for d in sub_dirs ]
        extra_options = list()
        cwd_parent = get_parent()
        if cwd_parent:
            extra_options.append('..')
        if cwd:
            extra_options.append('.')
            if allow_mkdir:
                extra_options.append('<Make new folder here>')

        # ask user
        next_dir = choose_from_list( 
            choices=extra_options + selection_list,
            default=extra_options.index('.') if '.' in extra_options else None 
        )
        
        if next_dir=='..':
            if cwd==cwd_parent:
                # happens when at the root of a drive on windows
                sub_dirs = windows_list_logical_drives()
                cwd = None
                continue
            cwd = cwd_parent
        elif next_dir=='.':
            return cwd
        elif next_dir=='<Make new folder here>':
            while True:
                new_dir_name = user_input(
                    prompt="New folder name",
                    accepted=lambda x: isinstance(x, str)
                )
                new_dir = cwd / make_FS_safe(new_dir_name.replace('.',''))
                if new_dir.is_dir():
                    print(f"'{new_dir_name}' already exists !")
                    continue
                new_dir.mkdir()
                cwd = new_dir
                break
        else:
            cwd = sub_dirs[ selection_list.index( next_dir ) ]
        sub_dirs = folder_get_subdirs( cwd )


def clear_screen() -> None:
    ''' Clears the terminal screen, multi-platform '''
    _os = Os()
    if _os.windows or _os.cygwin:
        os.system( 'CLS' )
    elif _os.linux or _os.wsl:
        os.system( 'clear' )


def skipNlines( n: int ) -> None:
    ''' Skips n lines in the terminal; used for vertical whitespace '''
    assert n>0, "drslib::skipNlines: Invalid number of lines entered : '{}'".format(n)
    print('{}'.format('\n'*(n-1)))
